from copy import deepcopy
from typing import Any, Dict, List, Optional, Union

import jwt

from services import errors, types
from services.security import scopes
from services.security.base import IAuth, ITokenStore
from services.security.utils import get_delta


class JWTAuth(IAuth):
    def __init__(self, conf: types.SecurityConfig, store: Optional[ITokenStore] = None):
        """
        It is a wrapper around jwt which produces jwt tokens.
        By default it will add a "exp" claim, other claims.

        Standard claims could be configurated from JWTConfing or when passing
        a payload to encode. In that case if both configurations exists, it will
        prioritize the payload configuration.
        """
        self.conf = conf.jwt
        self.store = store

    def get_username(self, request) -> Union[str, None]:
        if getattr(request.ctx, "token_data"):
            return request.ctx.token_data["usr"]
        return None

    async def generate_token(self, obj: types.JWTPayload) -> types.JWTResponse:
        encoded = self.encode(obj.custom, exp=obj.exp, iss=obj.iss, aud=obj.aud)
        rfk = None
        if self.store:
            rfk = await self._store_refresh_token(obj.custom["usr"])

        return types.JWTResponse(access_token=encoded, refresh_token=rfk)

    def validate_request(self, request, policies: List[str] = None, require_all=True):
        token = request.token
        if token:
            try:
                decoded = self.validate(token, policies, require_all)
                request.ctx.token_data = decoded
                if decoded.get("usr"):
                    request.ctx.username = decoded.get("usr")
                request.ctx.is_authenticated = True
                request.ctx.is_authorized = True
            except errors.AuthValidationFailed:
                pass

    def encode(self, payload: Dict[str, Any], exp=None, iss=None, aud=None) -> str:
        """Encode a payload into a JWT Token.

        Some standards claims like exp, iss and aud could be
        overwritten using this params.

        :param payload: a dictionary with any k/v pairs to add
        :param exp:  the “exp” (expiration time) claim identifies the expiration
        time on or after which the JWT MUST NOT be accepted for processing.
        if date or int is given it will overwrite the default configuration.
        :param iss: the “iss” (issuer) claim identifies the principal that issued the JWT.
        :param aud: The “aud” (audience) claim identifies the recipients that the JWT
        is intended for.
        """
        _secret = self._get_secret_encode()

        final = self._build_payload(payload, exp, iss, aud)

        encoded = jwt.encode(
            final,
            _secret,
            algorithm=self.conf.alg,
        )
        return encoded

    def decode(
        self, encoded, verify_signature=True, verify_exp=True, iss=None, aud=None
    ) -> Dict[str, Any]:
        _secret = self._get_secret_decode()

        _iss = iss or self.conf.issuer
        _aud = aud or self.conf.audience

        return jwt.decode(
            encoded,
            _secret,
            options={
                "verify_signature": verify_signature,
                "verify_exp": verify_exp,
                "require": self.conf.requires_claims,
            },
            aud=_aud,
            iss=_iss,
            algorithms=[self.conf.alg],
        )

    def validate(
        self,
        token: str,
        required_scopes: Optional[List[str]] = None,
        require_all=True,
        iss=None,
        aud=None,
    ) -> Dict[str, Any]:
        try:
            decoded = self.decode(token, iss=iss, aud=aud)
            if required_scopes:
                user_scopes: List[str] = decoded["scopes"]
                valid = scopes.validate(
                    required_scopes, user_scopes, require_all=require_all
                )
                if not valid:
                    raise errors.AuthValidationFailed()
        except jwt.InvalidTokenError as exc:
            raise errors.AuthValidationFailed() from exc

        return decoded

    def _get_secret_encode(self):
        """because jwt allows the use of a simple secret or a pairs of keys
        this function will look at the configuration to determine a secret to be used
        """
        if self.conf.keys:
            _secret = self.conf.keys.private
        else:
            _secret = self.conf.secret
        return _secret

    def _get_secret_decode(self):
        if self.conf.keys:
            _secret = self.conf.keys.public
        else:
            _secret = self.conf.secret
        return _secret

    def _build_payload(self, payload: Dict[str, Any], exp=None, iss=None, aud=None):
        _payload = deepcopy(payload)
        if not exp:
            exp = get_delta(self.conf.exp_min)
        else:
            exp = get_delta(exp)
        _iss = iss or self.conf.issuer
        if _iss:
            _payload.update({"iss": _iss})

        _aud = aud or self.conf.audience
        if _aud:
            _payload.update({"aud": _aud})

        _payload.update({"exp": exp})
        return _payload

    def update_response(self, request, response):
        pass

    async def _store_refresh_token(self, username: str, ttl=None) -> str:
        refresh = self.store.generate()
        _ttl = ttl or self.conf.ttl_refresh_token
        await self.store.put(f"{username}.{refresh}", username, ttl=_ttl)

        return refresh

    async def _validate_refresh_token(self, access_token, refresh_token) -> bool:
        is_valid = False
        decoded = self.decode(access_token, verify_exp=False)
        value = await self.store.get(f"{decoded['usr']}.{refresh_token}")
        if decoded["usr"] == value:
            is_valid = True
        return is_valid

    async def refresh_token(self, access_token, refresh_token) -> types.JWTResponse:
        if self.store:
            is_valid = await self._validate_refresh_token(access_token, refresh_token)
            if is_valid:
                decoded = self.decode(access_token, verify_exp=False)
                await self.store.delete(f"{decoded['usr']}.{refresh_token}")

                _new_refresh = await self._store_refresh_token(decoded["usr"])
                _new_tkn = self.encode(decoded)
                new_jwt = types.JWTResponse(
                    access_token=_new_tkn, refresh_token=_new_refresh
                )
                return new_jwt
        raise errors.AuthValidationFailed()





class CommandExecutionException(Exception):
    def __init__(self, message):
        super().__init__(message)


class BadConfigurationException(Exception):
    def __init__(self, bad_param):
        super().__init__(f"{bad_param} is badly used")


class UserNotFound(Exception):
    def __init__(self, username):
        super().__init__(f"User: {username} not found")


class GroupNotFound(Exception):
    def __init__(self, group):
        super().__init__(f"Group: '{group}' not found")


class DBObjectNotFound(Exception):
    def __init__(self, table, key):
        super().__init__(f"{key} not found in table {table}")


class AuthValidationFailed(Exception):
    pass



from sanic import Blueprint, Request
from sanic.response import json
{% if data.tasks -%}
from services.workers import get_queue
from {{ data.app_name }}.tasks import TaskExample
{% endif -%}
{% if data.storage -%}
from services.storage import Storage
from services.server import stream_reader
{% endif -%}

{{ data.app_name }}_bp = Blueprint("{{ data.app_name }}", url_prefix="{{ data.app_name }}", version="v1")


@{{ data.app_name }}_bp.get("/")
async def json_handler(request: Request):
    return json(dict(msg="hello wolrd! from app {{ data.app_name }}"), 200)


{% if data.tasks -%}
@{{ data.app_name }}_bp.post("/worker")
async def worker_handler(request: Request):
    t = TaskExample(do="web", wait=5)
    q = get_queue(request, qname="default")
    task = await q.submit(name="dummy", params=t.dict())
    return json(task.dict(exclude={'created_at', 'updated_at'}), 201)
{% endif -%}

{% if data.storage -%}
@{{ data.app_name }}_bp.post("/_/upload/<filename:str>", stream=True)
async def upload_example(request, filename, store: Storage):
    _store = store.get_storage(name="default")
    await _store.put_stream(filename, stream_reader(request))
    return json(dict(msg="ok"), 200)
{% endif -%}



from abc import ABC, abstractmethod
from importlib import import_module
from typing import Callable, List, Optional, Tuple

from pydantic import BaseModel
from sanic import HTTPResponse, Request, Sanic

from services.db import AsyncSQL
from services.security import Authenticator, IAuth
from services.types import Settings


def _init_blueprints(app, blueprints_allowed, package_dir="services.web"):
    """
    It will import bluprints from modules that ends with "_bp" and belongs
    to the package declared in `package_dir`
    """
    blueprints = set()
    mod = app.__module__
    for mod_name in blueprints_allowed:
        modules = import_module(f"{package_dir}.{mod_name}", mod)
        for el in dir(modules):
            if el.endswith("_bp"):
                bp = getattr(modules, el)
                blueprints.add(bp)

    for bp in blueprints:
        print("Adding blueprint: ", bp.name)
        app.blueprint(bp)


class ViewSet(BaseModel):
    blueprints: List[str]
    package: str


class WebAppSpec(ABC):
    """Spec whichs define a WebApp.
    A WebApp can have endpoints, or not.
    """

    name: str
    views: List[ViewSet]
    auth: Authenticator

    #    self.auth: Authenticator = Authenticator()
    def __init__(self, app: Sanic, settings: Settings):
        self.auth: Authenticator = Authenticator()
        app.ctx.auth_beta = self.auth
        self.app = app
        self.init(app, settings)

    @abstractmethod
    def init(self, app: Sanic, settings: Settings):
        pass

    def register_auth_validator(self, app: Sanic, name: str, auth: IAuth):
        app.ext.dependency(auth)
        self.auth.register_validator(name, auth)

    def init_blueprints(self, app: Sanic):
        for view in self.views:
            _init_blueprints(app, view.blueprints, package_dir=view.package)

    def get_db(self, app: Sanic, name="default") -> AsyncSQL:
        return app.ctx.databases[name]

    def get_app_settings(self, app: Sanic) -> Settings:
        return app.config.SETTINGS

    def get_request_settings(self, request: Request) -> Settings:
        return request.app.config.SETTINGS

    def get_request_auth(self, request: Request) -> Authenticator:
        return request.app.ctx.auth_beta



import toml


def get_version(poetry="pyproject.toml") -> str:
    """Get the version of the package from pyproject file"""
    with open("pyproject.toml", "r", encoding="utf-8") as f:
        data = toml.loads(f.read())

    return data["tool"]["poetry"]["version"].strip()


if __name__ == "__main__":
    print(get_version())



from abc import ABCMeta, abstractmethod
from secrets import token_urlsafe
from typing import Generic, List, Optional, TypeVar, Union
from pydantic import BaseModel

from services import types

StoreDriverT = TypeVar("StoreDriverT")
UserAuthT = TypeVar("UserAuthT", bound=BaseModel)


class ITokenStore(Generic[StoreDriverT], metaclass=ABCMeta):
    def __init__(self, conf: types.SecurityConfig, driver: StoreDriverT):
        self.conf = conf
        self.driver = driver

    @classmethod
    @abstractmethod
    async def from_conf(
        cls, conf: types.SecurityConfig, driver: Optional[StoreDriverT] = None
    ) -> "ITokenStore":
        pass

    @abstractmethod
    async def put(self, key: str, value: str, ttl: Optional[int] = None) -> bool:
        pass

    @abstractmethod
    async def get(self, key: str) -> Union[str, None]:
        pass

    @abstractmethod
    async def delete(self, key: str):
        pass

    @staticmethod
    def generate(sign: Optional[str] = None) -> str:
        return token_urlsafe(16)


class IAuth(metaclass=ABCMeta):
    @abstractmethod
    def get_username(self, request) -> Union[str, None]:
        pass

    @abstractmethod
    def validate_request(self, request, policies: List[str] = None, require_all=True):
        pass

    @abstractmethod
    def update_response(self, request, response):
        pass



from datetime import datetime, timedelta

from services import types


def get_delta(delta_min: int) -> int:
    """Returns a timestamp addding a delta_min value to the utc now date."""
    delta = datetime.utcnow() + timedelta(minutes=delta_min)
    return int(delta.timestamp())


def open_keys(pub, priv) -> types.KeyPairs:
    with open(pub, "r") as f:
        pub = f.read()
        pub = pub.strip()
    with open(priv, "r") as f:
        priv = f.read()
        priv = priv.strip()
    return types.KeyPairs(public=pub, private=priv)



import importlib
from typing import Optional

from alembic import command
from alembic.config import Config as AlembicConfig
from services.utils import get_package_dir


def alembic_config(dburi: str,
                   *,
                   ini_section="alembic",
                   version_table: str,
                   # models_name: List[str],
                   migrations_pkg="services") -> AlembicConfig:
    """
    :param dburi: full dsn of the SQL connection
    :param version_table: table name used to register migrations versions
    :models_name: 
    """
    # dir_ = get_package_dir(migrations_pkg)
    # alembic_file = f"{dir_}/alembic.ini"
    # TODO: add base path
    alembic_file = "alembic.ini"
    # this is need for alembic to register models
    models_module = importlib.import_module(f"{migrations_pkg}.models")
    # models = [importlib.import_module(mod) for mod in models_name]

    alembic_cfg = AlembicConfig(alembic_file, ini_section=ini_section)
    # alembic_cfg.set_main_option("script_location", migrations_pkg)
    alembic_cfg.set_main_option("version_table", version_table)
    # alembic_cfg.set_main_option("target_metadata", version_table)
    alembic_cfg.set_main_option("sqlalchemy.url", dburi)
    return alembic_cfg


class Migration:
    def __init__(self, dburi, *, section: str,
                 alembic_file: str = "alembic.ini"):
        self.dburi = dburi
        self._file = alembic_file
        self._section = section

    def load_config(self) -> AlembicConfig:
        alembic_cfg = AlembicConfig(self._file, ini_section=self._section)
        alembic_cfg.set_main_option("sqlalchemy.url", self.dburi)
        # _models = alembic_cfg.get_main_option("models_module")
        # models_module = importlib.import_module(str(_models))
        return alembic_cfg

    def upgrade(self, to="head"):
        cfg = self.load_config()
        command.upgrade(cfg, to)

    def downgrade(self, to="head-1"):
        cfg = self.load_config()

        command.downgrade(cfg, to)

    def revision(self,
                 rev_id: Optional[str] = None,
                 message: Optional[str] = None,
                 sql: bool = False,
                 head: str = "head",
                 branch_label: Optional[str] = None,
                 version_path: Optional[str] = None,
                 depends_on: Optional[str] = None,
                 autogenerate=True
                 ):
        """
        :param rev_id: optional to specify a revision id
        :param message: to be used as message
        :param sql: to print in the stdout the sql generated
        :param head:  head revision to build the new revision upon as a parent
        :param depends_on:  optional list of "depends on" identifiers
        """

        cfg = self.load_config()
        command.revision(
            cfg,
            message=message,
            rev_id=rev_id,
            autogenerate=autogenerate,
            sql=sql,
            head=head,
            depends_on=depends_on,
            branch_label=branch_label,
            version_path=version_path,

        )



from secrets import token_urlsafe
from typing import Optional, Union, Dict
from services import types

from services.security.base import ITokenStore


class MemoryTokenStore(ITokenStore[dict]):
    def __init__(self, conf: types.SecurityConfig, driver={}):
        self.conf = conf
        self.data: Dict[str, str] = {}

    @classmethod
    async def from_conf(
        cls, conf: types.SecurityConfig, driver: Optional[dict] = None
    ) -> "MemoryTokenStore":
        if not driver:
            driver = {}
        return cls(conf, driver)

    async def put(self, key: str, value: str, ttl: Optional[int] = None) -> bool:
        self.data[key] = value
        return True

    async def get(self, key: str) -> Union[str, None]:
        return self.data.get(key)

    async def delete(self, key: str):
        if self.data.get(key):
            del self.data[key]

    @staticmethod
    def generate(sign: Optional[str] = None) -> str:
        return token_urlsafe(16)



from sqlalchemy.ext.compiler import compiles
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.types import BINARY, JSON, BigInteger

from services.db import SQL, AsyncSQL

# equivalent to Base = declarative_base()


class Base(DeclarativeBase):
    pass


async def async_create_all(db: AsyncSQL):
    await db.create_all(Base.metadata)


def create_all(db: SQL):
    db.create_all(Base.metadata)


@compiles(JSON, "postgresql")
def compile_jsonb_postgres(type_, compile, **kw):
    return "JSONB"


@compiles(BINARY, "postgresql")
def compile_bytea_postgres(type_, compiler, **kw):
    return "BYTEA"


@compiles(BigInteger, "sqlite")
def compile_integer_sqlite(type_, compiler, **kw):
    return "Integer"



from typing import Callable, Optional, Union

from itsdangerous import TimestampSigner
from itsdangerous.exc import BadSignature
from sqlalchemy import Column, MetaData, String, Table
from sqlalchemy import delete as sqldelete
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine

from services import types
from services.security.base import ITokenStore

meta = MetaData()


def _create_engine(uri) -> AsyncEngine:
    return create_async_engine(uri)


class SQLTokenStore(ITokenStore[AsyncEngine]):
    """
    TODO: set created time to validate expired token
    """

    tokens = Table(
        "refresh_tokens",
        meta,
        Column("key", String(16), primary_key=True),
        Column("value", String(16)),
    )

    def __init__(self, conf: types.SecurityConfig, driver: AsyncEngine):
        super().__init__(conf, driver)
        self.sign = TimestampSigner(conf.secret_key)

    async def _dispose(self):
        if ":memory:" not in self.conf.token_store_uri:
            await self.driver.dispose()

    @classmethod
    async def from_conf(
        cls, conf: types.SecurityConfig, driver: Optional[AsyncEngine] = None
    ) -> "SQLTokenStore":
        if not driver:
            driver = create_async_engine(conf.token_store_uri)
        async with driver.begin() as conn:
            await conn.run_sync(meta.create_all)

        if ":memory:" not in conf.token_store_uri:
            await driver.dispose()
        return cls(conf, driver)

    async def put(self, key: str, value: str, ttl: Optional[int] = None) -> bool:
        svalue = self.sign.sign(value).decode("utf-8")

        async with self.driver.begin() as conn:
            await conn.execute(self.tokens.insert(), [{"key": key, "value": svalue}])

        await self._dispose()
        return True

    async def get(self, key: str) -> Union[str, None]:
        final = None
        invalid = False
        async with self.driver.begin() as conn:
            result = await conn.execute(
                select(self.tokens).where(self.tokens.c.key == key).limit(1)
            )
            row = result.fetchall()
            if row:
                final = row[0][1]
                try:
                    final = self.sign.unsign(final, max_age=self.conf.ttl_refresh_token)
                    final = final.decode()
                except BadSignature:
                    invalid = True

        if invalid:
            final = None
            await self.delete(key)
        await self._dispose()
        return final

    async def delete(self, key: str):
        async with self.driver.begin() as conn:
            await conn.execute(sqldelete(self.tokens).where(self.tokens.c.key == key))

        await self._dispose()



from sqlalchemy.ext.mutable import Mutable

# from sqlalchemy.ext.compiler import compiles


# @compiles(JSON, "postgresql")
# def compile_jsonb_postgres(type_, compile, **kw):
#     return "JSONB"
#
#
# @compiles(BINARY, "postgresql")
# def compile_bytea_postgres(type_, compiler, **kw):
#     return "BYTEA"
#
#
# @compiles(BigInteger, "sqlite")
# def compile_integer_sqlite(type_, compiler, **kw):
#     return "Integer"
#

# @compiles(DECIMAL, "sqlite")
# def compile_integer_sqlite(type_, compiler, **kw):
#     return "Text"


class MutableList(Mutable, list):
    """This allow modify an ARRAY custom type
    From https://gist.github.com/kirang89/22d111737af0fca251e3

    """

    def __setitem__(self, key, value):
        list.__setitem__(self, key, value)
        self.changed()

    def __delitem__(self, key):
        list.__delitem__(self, key)
        self.changed()

    def append(self, value):
        list.append(self, value)
        self.changed()

    def pop(self, index=0):
        value = list.pop(self, index)
        self.changed()
        return value

    @classmethod
    def coerce(cls, key, value):
        if not isinstance(value, MutableList):
            if isinstance(value, list):
                return MutableList(value)
            return Mutable.coerce(key, value)
        else:
            return value



SANIC_APP = "srv"
BASE_PATH_ENV = "SRV_BASE_PATH"
LOGGER = "srv"
SETTINGS_MODULE = "server_conf.settings"
SETTINGS_MODULE_VAR = "SRV_SETTINGS_MODULE"
SUPER_USER_GROUP = "root"
GOOGLE_AUTH_ENV = "GOOGLE_APPLICATION_CREDENTIALS"



import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import TSVECTOR
from sqlalchemy.orm import declarative_mixin, declared_attr


def tsvector_column(fields="title || ' ' || description", lang="english"):
    return sa.Column(
        TSVECTOR,
        sa.Computed(f"to_tsvector('{lang}', {fields})", persisted=True),
    )


def search_stmt(model, text, field="tsvector"):
    col = getattr(model, field)
    stmt = sa.select(model).where(col.bool_op("@@")(sa.func.websearch_to_tsquery(text)))

    return stmt


def search_with_rank_stmt(model, text, field="tsvector"):
    """Returns a tuple with the model and the rank value

    check: https://www.postgresql.org/docs/current/textsearch-controls.html

    used as:

    .. code-block:: python


        stmt = search_with_rank_stmt(PersonModel, "google")
        with db.session() as s:
            res = s.execute(stmt)
            values = list(res)

    """
    col = getattr(model, field)
    stmt = sa.select(
        model, sa.func.ts_rank_cd(col, sa.func.websearch_to_tsquery(text)).label("rank")
    ).where(col.bool_op("@@")(sa.func.websearch_to_tsquery(text)))
    # stmt = sa.select(model.id, sa.func.ts_rank_cd(col, sa.func.websearch_to_tsquery(text), 0).label("rank"))
    # stmt = stmt.order_by(sa.desc(sa.func.ts_rank_cd(col, sa.func.websearch_to_tsquery(text))))
    stmt = stmt.order_by(sa.desc("rank"))

    return stmt


@declarative_mixin
class WithSearchField:
    # searcheable = sa.Column(TSVECTOR, sa.Computed)

    @declared_attr
    def __table_args__(cls):
        return (
            sa.Index(
                f"idx__{cls.__tablename__}__tsvector",
                "tsvector",
                postgresql_using="gin",
            ),
        )

    @classmethod
    def websearch(cls, text, field="tsvector"):
        col = getattr(cls, field)
        stmt = sa.select(cls).where(
            col.bool_op("@@")(sa.func.websearch_to_tsquery(text))
        )
        return stmt


class MixinSearch:
    @classmethod
    def fulltext_search(cls, search_string, field, lang="english"):
        return sa.select(cls).filter(
            sa.func.websearch_to_tsquery(lang, getattr(cls, field)).match(
                search_string, postgresql_regconfig=lang
            )
        )



import errno
import importlib.util
import os

from jinja2 import Environment, FileSystemLoader, select_autoescape


def get_package_dir(pkg):
    spec = importlib.util.find_spec(pkg)
    return spec.submodule_search_locations[0]


def render(filename, templates_dir=None, *args, **kwargs):
    path = (
        f"{get_package_dir('services')}/files"
        if templates_dir is None
        else templates_dir
    )
    env = Environment(
        loader=FileSystemLoader(path),
        autoescape=select_autoescape(),
    )
    tpl = env.get_template(filename)
    return tpl.render(*args, **kwargs)


def render_to_file(template, dst, templates_dir=None, *args, **kwargs):
    text = render(template, templates_dir, *args, **kwargs)
    with open(dst, "w", encoding="utf-8") as f:
        f.write(text)
    return text



from sanic import Blueprint, Request
from sanic.response import json
from services.workers import get_queue
from example.tasks import TaskExample
from services.storage import Storage
from services.utils import stream_reader
example_bp = Blueprint("example", url_prefix="example", version="v1")


@example_bp.get("/")
async def json_handler(request: Request):
    return json(dict(msg="hello wolrd! from app example"), 200)


@example_bp.post("/worker")
async def worker_handler(request: Request):
    t = TaskExample(do="web", wait=5)
    q = get_queue(request, qname="default")
    task = await q.submit(name="dummy", params=t.dict())
    return json(task.dict(exclude={'created_at', 'updated_at'}), 201)
@example_bp.post("/_/upload/<filename:str>", stream=True)
async def upload_example(request, filename, store: Storage):
    _store = store.get_storage(name="default")
    await _store.put_stream(filename, stream_reader(request))
    return json(dict(msg="ok"), 200)



import functools
import ssl
from dataclasses import dataclass
from typing import Any, Callable, Dict, Generator, List, Optional, Tuple, Union
from urllib.parse import urlparse

from pydantic.validators import make_arbitrary_type_validator
from redis.asyncio import ConnectionPool, Redis
from redis.asyncio.sentinel import Sentinel
from redis.exceptions import RedisError, WatchError


class SSLContext(ssl.SSLContext):
    """
    Required to avoid problems with
    """

    @classmethod
    def __get_validators__(cls) -> Generator[Callable[..., Any], None, None]:
        yield make_arbitrary_type_validator(ssl.SSLContext)


@dataclass
class RedisSettings:
    """
    No-Op class used to hold redis connection redis_settings.
    """

    host: str = "localhost"
    port: int = 6379
    database: int = 0
    username: Optional[str] = None
    password: Optional[str] = None
    ssl: Union[bool, None, SSLContext] = None
    conn_timeout: int = 1
    conn_retries: int = 5
    conn_retry_delay: int = 1
    decode_responses: bool = True

    sentinel: bool = False
    sentinel_master: str = "mymaster"

    @classmethod
    def from_dsn(cls, dsn: str, decode_responses=True) -> "RedisSettings":
        conf = urlparse(dsn)
        assert conf.scheme in {"redis", "rediss"}, "invalid DSN scheme"
        return RedisSettings(
            host=conf.hostname or "localhost",
            port=conf.port or 6379,
            ssl=conf.scheme == "rediss",
            username=conf.username,
            password=conf.password,
            database=int((conf.path or "0").strip("/")),
            decode_responses=decode_responses,
        )

    def __repr__(self) -> str:
        return "RedisSettings({})".format(
            ", ".join(f"{k}={v!r}" for k, v in self.__dict__.items())
        )


def create_pool(
    dsn_url: str,
    decode_responses=True,
) -> Redis:
    """
    Create a new redis pool, retrying up to ``conn_retries`` times if the connection fails.
    Returns a :class:`arq.connections.ArqRedis` instance, thus allowing job enqueuing.
    """
    settings = RedisSettings.from_dsn(dsn_url)

    assert not (
        type(settings.host) is str and settings.sentinel
    ), "str provided for 'host' but 'sentinel' is true; list of sentinels expected"

    if settings.sentinel:

        def pool_factory(*args: Any, **kwargs: Any) -> Redis:
            client = Sentinel(
                *args, sentinels=settings.host, ssl=settings.ssl, **kwargs
            )
            return client.master_for(settings.sentinel_master)

    else:
        pool_factory = functools.partial(
            Redis,
            host=settings.host,
            port=settings.port,
            socket_connect_timeout=settings.conn_timeout,
            ssl=settings.ssl,
            decode_responses=settings.decode_responses,
        )

    pool = pool_factory(
        db=settings.database,
        username=settings.username,
        password=settings.password,
        encoding="utf8",
    )
    return pool



from .simple_tags import StandaloneTag


class Static(StandaloneTag):
    tags = {"static"}

    def __init__(self, environment):
        super().__init__(environment)
        # environment.extend(vite_metadata_prefix="", vite_metadata=None)
        environment.extend(staticfiles_prefix="", staticfiles=None)

    def render(self, static_file):
        # static_dir = self.environment.staticfiles[static_key]
        static_dir = self.environment.staticfiles["public"]["uripath"]
        return f"{static_dir}{static_file}"



import sys
from pathlib import Path
from secrets import token_urlsafe

import click
from rich.console import Console

from services import init_script

console = Console()


@click.command()
@click.option(
    "--vite/--no-vite",
    default=False,
    is_flag=True,
    show_default=True,
    help="Get support for Vite, HTML feature should be enabled",
)
@click.option(
    "--html/--no-html",
    default=False,
    is_flag=True,
    show_default=True,
    help="Add a index view example and templates",
)
@click.option(
    "--web/--no-web",
    default=True,
    is_flag=True,
    show_default=True,
    help="Add sanic web server",
)
@click.option(
    "--users/--no-users",
    default=False,
    is_flag=True,
    show_default=True,
    help="Provide the user system",
)
@click.option(
    "--tasks/--no-tasks",
    default=False,
    is_flag=True,
    show_default=True,
    help="Add a dummy task and the worker as example",
)
@click.option(
    "--sql/--no-sql",
    default=False,
    is_flag=True,
    show_default=True,
    help="Add DB related files",
)
@click.option(
    "--storage/--no-storage",
    default=False,
    is_flag=True,
    show_default=True,
    help="Add a default storage",
)
@click.option(
    "--name",
    default=None,
    help="Name of the project, if empty it will ask you for a name",
)
@click.argument("base_path")
def create_service_project(
    base_path, web, name, vite, html, users, tasks, sql, storage
):
    """Start a new project"""
    root = Path(base_path).resolve()
    opts = init_script.ScriptOpts(
        base_path=root,
        secret_key=token_urlsafe(32),
        app_name=name,
        users=users,
        vite_enabled=vite,
        tasks=tasks,
        sql=sql,
        storage=storage,
        html=html,
        web=web,
    )
    print(opts)
    if opts.tasks and not opts.web:
        console.print("[red] To use tasks feature you should enable [cyan]web[/][/]")
        sys.exit(-1)
    if opts.vite_enabled and not opts.web:
        console.print("[red] To use vite feature you should enable [cyan]web[/][/]")
        sys.exit(-1)
    if opts.html and not opts.web:
        console.print("[red] To use html feature you should enable [cyan]web[/][/]")
        sys.exit(-1)
    init_script.create_project(opts)



import os
from functools import partial

import click
from services import defaults


@click.command(name="web")
@click.option("--host", "-H", default="127.0.0.1", help="Listening Host")
@click.option("--port", "-p", default="8000", help="Listening Port")
@click.option("--workers", "-w", default=1, help="How many workers start?")
@click.option(
    "--auto-reload", "-r", default=False, is_flag=True, help="Enable Auto reload"
)
@click.option(
    "--access-log", "-L", default=False, is_flag=True, help="Enable access_log"
)
@click.option(
    "--single", "-S", default=False, is_flag=True, help="Run in a single process"
)
@click.option(
    "--storage", "-s", default=True, is_flag=True, help="With storage enabled"
)
@click.option("--dev", "-d", default=False, is_flag=True, help="Dev mode")
@click.option(
    "--settings-module",
    "-s",
    default=defaults.SETTINGS_MODULE,
    help="Fullpath to settings module",
)
def web_cli(
    host, port, workers, auto_reload, access_log, single, dev, settings_module, storage
):
    """Run Web Server"""
    # pylint: disable=import-outside-toplevel
    from sanic import Sanic
    from sanic.worker.loader import AppLoader
    from services.conf import load_conf
    from services.server import create_srv

    # from playpy.web import sanic_init_indexer
    # from playpy.types import Settings as PlaySettings
    # pwd = os.getcwd()

    settings = load_conf(settings_module)

    host = host or settings.HOST
    port = port or settings.PORT
    w = int(workers)

    settings.SINGLE_PROCESS = single

    print(f"BASE_PATH: {settings.BASE_PATH}")
    print(f"SETTINGS_MODULE: {settings.SETTINGS_MODULE}")
    print(f"Single Process: {settings.SINGLE_PROCESS}")
    print(f"Dev mode: {dev}")
    print(f"Access log: {access_log}")
    print(f"Autoreload: {auto_reload}")
    print(f"Workers: {w}")
    print(f"Listening: {host}:{port}")
    print(f"OS PID: {os.getpid()}")
    print(f"With storage: {storage}")

    loader = AppLoader(factory=partial(create_srv, settings=settings))
    srv = loader.load()
    srv.prepare(
        host=host,
        port=int(port),
        dev=dev,
        workers=w,
        auto_reload=auto_reload if not single else False,
        access_log=access_log,
        single_process=single,
    )
    if single:
        Sanic.serve_single()
    else:
        Sanic.serve(primary=srv, app_loader=loader)

    # srv = create_srv(settings)
    # srv.run(
    #     host=host,
    #     port=int(port),
    #     access_log=access_log,
    #     workers=w,
    #     debug=debug,
    #     auto_reload=reload,
    # )



import sys

from services import defaults, types

# WARNING:
# We do our best effort to keep sensible information private
# but in the scenario of an intrusion into the network or machines
# where servers are hosted, then it represents a risk for the information
# stored in those machines.


# General Folders for the server



import asyncio
import inspect
import json
import logging
import traceback
from abc import ABC, abstractclassmethod, abstractmethod
from datetime import datetime
from enum import Enum
from functools import partial
from multiprocessing import Manager, Queue
from os import getpid
from queue import Empty
from typing import Any, Callable, Dict, List, Optional, Union

from pydantic import BaseModel, Field
from sanic import Sanic
from sanic.log import LOGGING_CONFIG_DEFAULTS, logger

from services.types import TasksBackend
from services.utils import get_class, get_function, secure_random_str

CTX_PREFIX = "queue_"
WORKER_PREFIX = "Queue-"


class QueueConfig(BaseModel):
    app_name: str
    qname: str = "default"
    backend: Optional[TasksBackend] = None


class TaskStatus(str, Enum):
    created = "CREATED"
    waiting = "WAITING"
    running = "RUNNING"
    cancelled = "CANCELLED"
    failed = "FAILED"
    done = "DONE"


class Task(BaseModel):
    name: str
    params: Dict[str, Any] = Field(default={})
    id: str = Field(default=secure_random_str())
    state: str = TaskStatus.created
    app_name: str = "test"
    timeout: int = 10
    result_ttl: int = 120
    created_at: datetime = Field(default=datetime.utcnow())
    updated_at: datetime = Field(default=datetime.utcnow())

    class Config:
        use_enum_values = True
        arbitrary_types_allowed = True


class _Task(BaseModel):
    task: Task
    future: asyncio.Task

    class Config:
        use_enum_values = True
        arbitrary_types_allowed = True


class IState(ABC):
    @abstractmethod
    async def add_task(self, task: Task):
        raise NotImplementedError()

    @abstractmethod
    async def get_task(self, taskid: str) -> Task:
        raise NotImplementedError()

    @abstractmethod
    async def update_status(self, taskid: str, status: str) -> bool:
        raise NotImplementedError()

    @abstractmethod
    async def list_tasks(self) -> List[Task]:
        raise NotImplementedError()

    @abstractmethod
    async def delete_task(self, taskid: str) -> bool:
        raise NotImplementedError()

    @abstractmethod
    async def get_result(self, taskid: str) -> Dict[str, Any]:
        raise NotImplementedError()

    @abstractmethod
    async def set_result(self, taskid: str, *, result: Dict[str, Any], status: str):
        raise NotImplementedError()

    @abstractclassmethod
    async def from_uri(cls, uri: str, extra: Dict[str, Any] = {}) -> "IState":
        raise NotImplementedError()

    @abstractmethod
    async def clean(self):
        raise NotImplementedError()

    @abstractmethod
    async def clean_failed(self) -> List[str]:
        raise NotImplementedError()


def elapsed_time_from_finish(task: Task):
    n = datetime.utcnow()
    return (n - task.updated_at).total_seconds()


def elapsed_time_from_start(task: Task):
    n = datetime.utcnow()
    return (n - task.created_at).total_seconds()


def _get_kwargs(task: Task, fn: Callable) -> Dict[str, BaseModel]:
    annot = list(fn.__annotations__.keys())
    kwargs = {}
    if annot:
        params_key = annot[0]
        params = fn.__annotations__[params_key](**task.params)
        kwargs = {params_key: params}
    return kwargs


def _get_function(base_package, task: Task) -> Callable:
    _name = task.name
    if "." not in _name:
        fn = get_function(f"{base_package}.tasks.{_name}")
    else:
        fn = get_function(_name)
    return fn


def _exec_task(base_package, task: Task):
    _name = task.name
    logger.info("Starting task %s", _name)
    fn = _get_function(base_package, task)
    kwargs = _get_kwargs(task, fn)
    result = fn(**kwargs)
    logger.info("Finished task %s", _name)
    return result


async def init_backend(conf: TasksBackend) -> IState:
    Cls: IState = get_class(conf.backend_class)
    backend = await Cls.from_uri(conf.uri)
    return backend


class TaskQueue:
    def __init__(
        self, queue: Queue, *, backend: Optional[IState] = None, conf: QueueConfig
    ) -> None:
        self.queue = queue
        self.qname = conf.qname
        self._app_name = conf.app_name
        self.backend = backend

    def send(self, task: Task) -> None:
        self.queue.put_nowait(task.json())

    def receive(self, wait=True) -> Dict[str, Any]:
        if wait:
            rsp = self.queue.get()
            data = json.loads(rsp)
        else:
            try:
                rsp = self.queue.get_nowait()
                data = json.loads(rsp)
            except Empty:
                return {}
        return data

    async def submit(
        self,
        *,
        name: str,
        params: Dict[str, Any],
        timeout: int = 60,
        result_ttl: int = 900,
        debug=False,
    ) -> Task:
        task = Task(
            name=name,
            params=params,
            app_name=self._app_name,
            timeout=timeout,
            result_ttl=result_ttl,
        )
        if self.backend:
            await self.backend.add_task(task)
        if debug:
            _exec_task(self._app_name, task)
        else:
            self.send(task)

        return task

    @classmethod
    def setup(cls, app: Sanic, conf: QueueConfig):
        @app.after_server_start
        async def _setup_taskqueue(app: Sanic):
            back = None
            if conf.backend:
                back = await init_backend(conf.backend)
            q = getattr(app.shared_ctx, f"{CTX_PREFIX}{conf.qname}")
            tq = cls(q, conf=conf, backend=back)
            setattr(app.ctx, f"{CTX_PREFIX}{conf.qname}", tq)

    @staticmethod
    def get_from_request(request, qname: str) -> "TaskQueue":
        q = getattr(request.app.ctx, f"{CTX_PREFIX}{qname}")
        return q


class Scheduler:
    STOP = "__STOP__"
    backend: Optional[IState] = None

    def __init__(
        self,
        queue: TaskQueue,
        loop,
        base_package,
        timeout=60,
        max_jobs=5,
        backend: Optional[TasksBackend] = None,
    ):
        self.queue = queue
        self._loop = loop
        self._base_package = base_package
        self.timeout = timeout
        # self.sem = asyncio.BoundedSemaphore(max_jobs)
        self.sem = None
        self._max_jobs = max_jobs
        self.tasks: Dict[str, _Task] = {}
        self._backend = backend
        self.backend: Optional[IState] = None

    async def add_task(self, task: Task):
        if self._backend:
            logger.info("Adding task %s [%s]", task.name, task.id)
            await self.backend.add_task(task)

    async def _update_status(self, task: Task, status: str):
        task.updated_at = datetime.utcnow()
        task.state = status
        if self.backend:
            await self.backend.update_status(task.id, status=status)

    async def _set_result(self, task: Task, *, result, status: str):
        task.updated_at = datetime.utcnow()
        task.state = status
        if self.backend:
            logger.info("BACKEND SET RESULT")
            await self.backend.set_result(task.id, result=result, status=status)

    async def _delete_task(self, task: Task):
        if self.backend and task.state == TaskStatus.done.value:
            await self.backend.delete_task(task.id)

    async def exec_task(self, task: Task):
        # ctx = get_context("fork")
        # with concurrent.futures.ProcessPoolExecutor(mp_context=ctx) as pool:
        logger.info("Executing task %s [%s]", task.name, task.id)
        result = None
        fn = _get_function(self._base_package, task)
        kwargs = _get_kwargs(task, fn)
        status = TaskStatus.running.value
        result = None
        try:
            await self._update_status(task, status)
            if inspect.iscoroutinefunction(fn):
                result = await fn(**kwargs)
            else:
                result = await self._loop.run_in_executor(None, partial(fn, **kwargs))
            status = TaskStatus.done.value
        except asyncio.exceptions.TimeoutError as e:
            err = traceback.format_exc()
            logger.error("Task timeout error %s [%s]: %s", task.name, task.id, e)
            result = {"error": err}
            status = TaskStatus.failed.value
        except Exception as e:
            err = traceback.format_exc()
            logger.error("Task error %s [%s]: %s", task.name, task.id, e)
            result = {"error": err}
        finally:
            logger.info("SETTING RESULT")
            await self._set_result(task, result=result, status=status)

        return result

    def start_task(self, task: Task) -> asyncio.Task:
        _task = self._loop.create_task(self.exec_task(task))
        self.tasks[task.id] = _Task(task=task, future=_task)
        return _task

    async def _sentinel(self):
        logger.debug("> Cleaning")
        to_delete = []
        for k, task in self.tasks.items():
            if task.future.done():
                to_delete.append(k)
        for x in to_delete:
            del self.tasks[x]
        if self.backend:
            await self.backend.clean()

    async def init_backend(self):
        if self._backend:
            Cls: IState = get_class(self._backend.backend_class)
            self.backend = await Cls.from_uri(self._backend.uri)

    async def run(self):
        logger.info("> Starting scheduler")
        await self.init_backend()
        print("BACKEND CONF: ", self._backend)
        print("BACKEND OBJ: ", self.backend)
        sem = asyncio.Semaphore(self._max_jobs)
        while True:
            async with sem:
                task_dict = self.queue.receive(wait=False)
                if not task_dict:
                    await self._sentinel()
                    await asyncio.sleep(0.8)
                else:
                    task = Task(**task_dict)
                    await sem.acquire()
                    _task = self.start_task(task)
                    logger.info("task %s [%s] added", task.name, task.id)
                    _task.add_done_callback(lambda _: sem.release())

    def finish_pending_tasks(self):
        if self.tasks:
            tasks = [t.future for t in self.tasks.values()]
            drain = asyncio.wait_for(asyncio.gather(*tasks), timeout=60)
            self._loop.run_until_complete(drain)


def cpu_worker(name, queue: Queue, conf: QueueConfig, max_jobs=1) -> None:
    """based on https://amhopkins.com/background-job-worker"""
    logging.config.dictConfig(LOGGING_CONFIG_DEFAULTS)
    pid = getpid()
    logger.debug("max_jobs=%s unused variable", max_jobs)
    logger.info(">> CPU Bound worker reporting for duty: %s [%s]", name, pid)
    tq = TaskQueue(queue, conf=conf)

    try:
        while True:
            task_dict = tq.receive()
            task = Task(**task_dict)
            _exec_task(conf.app_name, task)

    except KeyboardInterrupt:
        logger.info("Shutting down %s", pid)
    finally:
        logger.info("Stopping CPU bound worker [%s]. Goodbye", pid)


def io_worker(name: str, queue: Queue, conf: QueueConfig, max_jobs=5) -> None:
    """based on https://amhopkins.com/background-job-worker"""
    logging.config.dictConfig(LOGGING_CONFIG_DEFAULTS)
    pid = getpid()
    logger.info(">> IO Bound worker reporting for duty: %s [%s]", name, pid)
    tq = TaskQueue(queue, conf=conf)
    loop = asyncio.new_event_loop()
    scheduler = Scheduler(
        tq,
        loop,
        base_package=conf.app_name,
        max_jobs=max_jobs,
        backend=conf.backend,
    )

    try:
        loop.create_task(scheduler.run())
        loop.run_forever()
        # loop.run_until_complete(scheduler.run())
    except KeyboardInterrupt:
        logger.info("Shutting down %s", pid)
    finally:
        scheduler.finish_pending_tasks()
    logger.info("Stopping IO bound worker [%s]. Goodbye", pid)


def standalone_cpu_worker(conf: QueueConfig, task: Task):
    logging.config.dictConfig(LOGGING_CONFIG_DEFAULTS)
    pid = getpid()
    logger.info(">> CPU Bound worker reporting for duty: %s", pid)
    loop = asyncio.new_event_loop()
    result = None
    status = task.state
    if conf.backend:
        backend = loop.run_until_complete(init_backend(conf.backend))
        loop.run_until_complete(backend.add_task(task))
    try:
        result = _exec_task(conf.app_name, task)
        status = TaskStatus.done.value

    except KeyboardInterrupt:
        result = {"error": "cancelled task"}
        status = TaskStatus.failed.value
        logger.error("Task [%s] cancelled", task.id)
        logger.info("Shutting down %s", pid)
    except Exception as e:
        result = {"error": str(e)}
        status = TaskStatus.failed.value
        logger.error("Task [%s] failed, msg: %s", task.id, e)
    finally:
        if conf.backend:
            loop.run_until_complete(
                backend.set_result(task.id, result=result, status=status)
            )

        logger.info("Stopping CPU bound worker [%s]. Goodbye", pid)


def standalone_io_worker(conf: QueueConfig, task: Task):
    logging.config.dictConfig(LOGGING_CONFIG_DEFAULTS)
    pid = getpid()
    logger.info(">> IO Bound worker reporting for duty: %s", pid)
    loop = asyncio.new_event_loop()
    tq = TaskQueue(Queue(), conf=conf)
    scheduler = Scheduler(tq, loop, base_package=conf.app_name, backend=conf.backend)
    if conf.backend:
        loop.run_until_complete(scheduler.init_backend())
        loop.run_until_complete(scheduler.add_task(task))
    try:
        # loop.create_task(scheduler.exec_task(task))
        loop.run_until_complete(scheduler.exec_task(task))
    except KeyboardInterrupt:
        logger.info("Shutting down %s", pid)
    finally:
        scheduler.finish_pending_tasks()
    logger.info("Stopping IO bound worker [%s]. Goodbye", pid)


def _get_queue_from_app(app, qname) -> Union[Queue, None]:
    try:
        q = getattr(app.shared_ctx, f"{CTX_PREFIX}{qname}")
    except AttributeError:
        return None
    return q


def create(
    app: Sanic,
    conf: QueueConfig,
    max_jobs: int = 1,
    jobs_per_worker: int = 5,
    wk: Callable = io_worker,
) -> None:
    @app.main_process_start
    async def start(app: Sanic):
        manager = Manager()
        if not _get_queue_from_app(app, conf.qname):
            q = manager.Queue()
            setattr(app.shared_ctx, f"{CTX_PREFIX}{conf.qname}", q)
        # app.shared_ctx.queue = manager.Queue()

    @app.main_process_ready
    async def ready(app: Sanic):
        q = _get_queue_from_app(app, qname=conf.qname)
        for ix in range(0, max_jobs):
            _name = f"{WORKER_PREFIX}{conf.qname}-{ix}"
            app.manager.manage(
                _name,
                wk,
                {
                    "name": _name,
                    "queue": q,
                    "conf": conf,
                    "max_jobs": jobs_per_worker,
                },
            )


def get_queue(request, qname="default") -> TaskQueue:
    q = getattr(request.app.ctx, f"{CTX_PREFIX}{qname}")
    return q



from .jwtauth import JWTAuth
from .web import Authenticator, get_authenticator, protected
from .base import IAuth, ITokenStore
from .password import PasswordScript

__all__ = ["JWTAuth", "Authenticator","get_authenticator", "protected", "JWTAuth", "IAuth", "ITokenStore", "PasswordScript"]



import os
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, BaseSettings, Field, RedisDsn

from services import defaults


class StaticDir(BaseModel):
    name: str
    uripath: str
    localdir: str


class Database(BaseModel):
    """
    Database configuration object to be used with
    :class:`services.db.SQL` and
    :class:`services.db.AsyncSQL`

    Examples:

    .. highlight:: python
    .. code-block:: python

        from services.types import Database
        dbconf = Database(uri="sqlite:////tmp/test.db")


    :param uri: full dsl url sqlite:////tmp/test.db
    :type uri: str
    :param name: optional value to identify a database
    :type name: str
    :param pool_size: size of the pool
    :type pool_size: int
    :param max_overflow: TODO
    :type max_overflow: int
    :param debug: maps with echo param in sqlalachemy
    :param description: optional value for metadata info about the database

    """

    async_url: Optional[str] = None
    sync_url: Optional[str] = None
    name: str = "default"
    pool_size: int = 20
    max_overflow: int = 0
    debug: bool = False  # debug
    # metadata
    description: Optional[str] = None


class MigrationType(BaseModel):
    models: List[str]
    package_dir: str
    version_table: str


class TasksBackend(BaseModel):
    uri: str = "sqlite+aiosqlite:///tasks.db"
    backend_class: str = "services.ext.sql.workers.SQLBackend"


class Storage(BaseModel):
    bucket: str = ".storage"
    store_class: str = "services.storage.AsyncLocal"


class SecuritySettings(BaseSettings):
    JWT_PUBLIC: Optional[str] = None
    JWT_PRIVATE: Optional[str] = None
    JWT_SECRET: Optional[str] = None
    JWT_ALG: str = "ES512"
    JWT_EXP: int = 30  # 30 minutes
    JWT_CLAIMS_REQUIRED: List[str] = ["exp"]
    JWT_ISS: Optional[str] = None
    JWT_AUD: Optional[str] = None
    REFRESH_TOKEN_TTL: int = 3600 * 168  # 7 days
    AUTH_SALT: str = "changeit"
    AUTH_ALLOW_REFRESH: bool = True
    AUTH_CLASS = "services.security.authentication.Auth"

    class Config:
        env_prefix = "SRV_"


class ViteConfig(BaseModel):
    """
    Vite configuration to be integrated into the backend.
    See: https://vitejs.dev/guide/backend-integration.html

    as an example, a base configuration in the front could be:

    .. code-block:: javascript

        export default defineConfig({
            root: resolve('./src'),
            base: '/',
            plugins: [svelte()],
            build: {
                manifest: true,
                outDir: resolve('./dist/vite'),
                emptyOutDir: true,
                target: 'es2015',
                rollupOptions: {
                input: {
                    main: resolve('./src/main.js'),
                        // explore: resolve('./theme/static/src/explore/main.js'),

                }
                }
            }
            })



    where the the python app is in: `test_app/test_app`
    and the frontend is in `test_app/front`
    With that in mind then:

    :param VITE_STATIC_URL_PATH: Where dynamic assets in vite are configurated.
    :param VITE_STATIC_DIR: Local path from the python app where the vite
        dynamic assets lives, in our example: "front/src/assets"
    :param VITE_OUTPUT_DIR: Where Vite generates the build,
        in our case: "front/dist/"
    :param VITE_DEV_SERVER: Where vite dev server is listening.
        By default it listen on "http://localhost:5173"
    :param VITE_DEV_MODE: if dev mode is true, the it will render the hmr script tag
    :param VITE_REACT_MODE: Enable if you are developing with REACT
    :param VITE_BASE: Base url in the vite server. Is  the same as
        https://vitejs.dev/config/server-options.html#server-base
        by default: "/"

    """

    # Where dynamic assets in vite are configurated
    VITE_STATIC_URL_PATH: str = "assets"
    # Local path from the python app where the vite dynamic assets are
    VITE_STATIC_DIR: str = "front/src/assets"
    # where vite put the buil
    VITE_OUTPUT_DIR: str = "fron/dist"
    # where vite dev server is listening
    VITE_DEV_SERVER: str = "http://localhost:5173"
    # if dev mode is true, the it will render the hmr script tag
    VITE_DEV_MODE: bool = True
    # special case for react
    VITE_REACT_MODE: bool = False
    # base url in the vite server the same as
    #    https://vitejs.dev/config/server-options.html#server-base
    VITE_BASE: str = "/"


class KeyPairs(BaseModel):
    public: str
    private: str


class JWTPayload(BaseModel):
    exp: Optional[int] = None
    iat: Optional[int] = None
    iss: Optional[str] = None
    sub: Optional[str] = None
    aud: Optional[str] = None
    custom: Dict[str, Any] = Field(default={})


class JWTConfig(BaseModel):
    alg: str
    exp_min: int = 30
    keys: Optional[KeyPairs] = None
    secret: Optional[str] = None
    issuer: Optional[str] = None
    audience: Optional[str] = None
    requires_claims: List[str] = ["exp"]
    ttl_refresh_token: int = 3600 * 168  # 7 days
    allow_refresh_token: bool = False

    class Config:
        extra = "forbid"


class SecurityConfig(BaseModel):
    secret_key: str
    jwt: Optional[JWTConfig] = None
    session: Optional[str] = None
    token_store_uri: str = "sqlite+aiosqlite:///:memory:"
    authenticators: List[str] = Field([])
    ttl_refresh_token: int = 3600 * 168  # 7 days
    domain: str = "localhost"


class Settings(BaseSettings):
    BASE_PATH: Union[str, Path]
    HOST: str = "localhost"
    PORT: str = "8000"
    DATABASES: Dict[str, Database] = {}
    REDIS: Optional[RedisDsn] = None
    REDIS_POOL_SIZE: int = 10
    TEMPLATES_DIR: List[str] = []
    TEMPLATES_PACKAGE_NAME: Optional[str] = None
    DEV_MODE: bool = False
    SECURITY: SecurityConfig = SecurityConfig(secret_key=os.urandom(16).hex())
    AUTHENTICATORS: List[str] = []
    CUSTOM_COMMANDS: List[str] = []
    USER_ENDPOINTS: bool = True
    USER_DB: str = "default"
    COMMANDS: List[str] = []
    TASKS: Optional[TasksBackend] = None

    APPS: List[str] = []

    # web
    STATIC_URL: str = ""
    STATICFILES_DIRS: List[StaticDir] = Field([])
    VITE_ENABLED: bool = False
    VITE_CONFIG: ViteConfig = ViteConfig()
    # logs
    LOGLEVEL: str = "INFO"
    LOGCONFIG: Dict[str, Any] = {}
    DEBUG: bool = False

    STORAGE: Optional[Dict[str, Storage]] = None

    # MIGRATIONS: Dict[str, Migration] = {}

    CORS_ORIGINS: Union[List[str], str] = "*"
    CORS_ALLOW_HEADERS: Union[List[str], str] = "*"
    SANIC_APP_NAME: str = defaults.SANIC_APP
    SINGLE_PROCESS: bool = False
    SETTINGS_MODULE: Optional[str] = None

    class Config:
        env_prefix = "SRV_"
        smart_union = True


class UserLogin(BaseModel):
    username: str
    password: str

    class Config:
        extra = "forbid"


class JWTResponse(BaseModel):
    access_token: str
    refresh_token: Optional[str] = None


class TokenCreds(BaseModel):
    access_token: str
    refresh_token: Optional[str] = None

    class Config:
        extra = "forbid"


class DefaultMeta(BaseModel):
    keywords: Optional[str] = None
    description: Optional[str] = None
    author: Optional[str] = None
    viewport: str = "width=device-width, initial-scale=1.0, maximum-scale=1"


class OGMeta(BaseModel):
    title: str
    description: str
    url: str
    image: str
    type_: str = "website"


class HtmlData(BaseModel):
    # navbar: List[MenuOption]
    ctx: Dict[str, Any]
    title: str
    content: Dict[str, Any] = Field({})
    lang: str = "en"
    meta: DefaultMeta = Field(DefaultMeta())
    og: Optional[OGMeta] = None


class Cookie(BaseModel):
    """
    expires: datetime - The time for the cookie to expire on the client’s browser.

    :param max_age:  Number of seconds the cookie should live for.
    :param path: The subset of URLs to which this cookie applies. Defaults to /.
    :param domain: Specifies the domain for which the cookie is valid.
        An explicitly specified domain must always start with a dot.
    :param max_age: Number of seconds the cookie should live for.
    :param secure: Specifies whether the cookie will only be sent
        via HTTPS.
    :param httponly: Specifies whether the cookie cannot be read by JavaScript.
    :param samesite: Default is browser dependent, specification states
        (Lax, Strict, and None) are valid values.
    """

    value: str
    domain: str
    httponly: bool = True
    path: str = "/"
    secure: bool = True
    samesite: str = "lax"
    max_age: int = Field(alias="max-age", default=60 * 60)


class Bucket(BaseModel):
    name: str
    url: str
    storage_class: str
    location: str
    versioning: bool = False
    labels: Optional[Dict[str, str]] = None
    public: bool = False
    created_at: Optional[datetime] = None

    def __repr__(self):
        return f"<Bucket {self.name}>"

    def __str__(self):
        return f"<Bucket {self.name}>"



# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
from typing import Any, Callable, Dict, List, Optional, Type, Union

from jinja2 import (
    Environment,
    FileSystemLoader,
    PackageLoader,
    Template,
    select_autoescape,
)
from jinja2.ext import Extension
from sanic import Sanic
from services.types import Settings
from services.utils import open_json

from .static import Static
from .vite import ViteAsset, ViteDev


class Render:
    # inspired by https://community.sanicframework.org/t/using-jinja2-with-sanic/615
    # also see
    # https://github.com/dldevinc/jinja2-simple-tags

    def __init__(
        self,
        package_name: Optional[str] = None,
        folder_name: str = "templates",
        searchpath: Union[str, List] = [],
        enable_async=True,
        encoding="UTF-8",
        extensions=[],
    ):
        """
        Example:
        ``` loader = PackageLoader("project.ui", "pages")```
        :param package_name: Import name of the package that contains the template directory.
        :param folder_name: Folder inside of the package which contains the templates
        :param searchpath:  if provided then a FilesystemLoader will be used instead
        :param enable_async: for async environments, it will return awaitables.
        :param encoding: utf-8 by default
        :param extensions: add extensions to jinja, like vite dev server ext
        """
        self._package = package_name
        self._folder = folder_name
        self.encoding = encoding
        self._loader = self.build_loader(package_name, searchpath)
        self.env: Environment = Environment(
            loader=self._loader,
            autoescape=select_autoescape(),
            enable_async=enable_async,
            extensions=extensions,
        )

    def build_loader(self, package_name, searchpath):
        if package_name:
            loader = PackageLoader(
                package_name=self._package,
                package_path=self._folder,
                encoding=self.encoding,
            )
        elif searchpath:
            loader = FileSystemLoader(
                searchpath,
                encoding=self.encoding,
            )
        else:
            raise AttributeError("searchpath or package_name should be set")
        return loader

    def add_func(self, name: str, func: Callable):
        """it allows to add custom functions"""
        self.env.globals[name] = func

    def add_extension(self, ext: Union[str, Type[Extension]]):
        """
        To add an extension pass a list of extension classes or import paths
        to the extensions parameter of the Environment constructor
        """
        self.env.add_extension(ext)

    def add_static(self, app: Sanic, settings: Settings):
        self.add_extension(Static)
        self.env.static_url = settings.STATIC_URL
        self.env.staticfiles = settings.STATICFILES_DIRS

    def add_vite(self, app: Sanic, settings: Settings):
        conf = settings.VITE_CONFIG
        dev_server = f"{conf.VITE_DEV_SERVER}"
        if conf.VITE_BASE != "/":
            dev_server = f"{conf.VITE_DEV_SERVER}{conf.VITE_BASE}"
        self.add_extension(ViteDev)
        self.add_extension(ViteAsset)
        self.env.vite_dev_server = dev_server
        self.env.vite_dev_mode = conf.VITE_DEV_MODE
        self.env.vite_react_mode = conf.VITE_REACT_MODE

        if conf.VITE_DEV_MODE:
            static_dir = f"{settings.BASE_PATH}/{conf.VITE_STATIC_DIR}"
            app.static(conf.VITE_STATIC_URL_PATH, static_dir, name="vite_dev")
        else:
            static_dir = f"{settings.BASE_PATH}/{conf.VITE_OUTPUT_DIR}{conf.VITE_STATIC_URL_PATH}"
            app.static(conf.VITE_STATIC_URL_PATH, static_dir, name="vite_manifest")
            manifest_json = open_json(
                f"{settings.BASE_PATH}/{conf.VITE_OUTPUT_DIR}/manifest.json"
            )
            self.env.vite_manifest = manifest_json

    def init_app(self, app: Sanic):
        """
        add render as part of the context and add url_for to be used
        inside a tempalte
        """
        app.ctx.render = self
        self.env.globals.update(url_for=app.url_for)

    def _get_app(self, request) -> Sanic:
        return request.app

    async def async_render(self, request, tpl_name, **kwargs) -> str:
        template = self.env.get_template(tpl_name)
        # ctx = {
        #     "request": request.ctx.__dict__,
        # }
        app = self._get_app(request)
        conf = {}
        if app.config.SETTINGS.DEBUG:
            conf = app.config.SETTINGS
        rendered = await template.render_async(conf=conf, **kwargs)
        return rendered

    async def arender(self, tpl_name, **kwargs) -> str:
        template = self.env.get_template(tpl_name)
        # ctx = {
        #     "request": request.ctx.__dict__,
        # }
        rendered = await template.render_async(**kwargs)
        return rendered

    def render(self, tpl_name, **kwargs) -> str:
        template = self.env.get_template(tpl_name)
        # ctx = {
        #     "request": request.ctx.__dict__,
        # }
        rendered = template.render(**kwargs)
        return rendered

    def get_template(self, tpl_name) -> Template:
        return self.env.get_template(tpl_name)



from .migration import Migration
from .pages import (
    NextPage,
    Pagination,
    get_total,
    get_total_async,
    get_total_table,
    get_total_table_async,
)
from .helpers import (
    SQL,
    AsyncSQL,
    sqlite_async_uri,
    vacuum,
    async_vacuum,
    async_set_pragma,
    set_pragma,
    commit_or_rollback,
    acommit_or_rollback,
)
from .managers2 import GenericManager



import math
from dataclasses import dataclass

from sqlalchemy import Table, func, select
from sqlalchemy.sql.selectable import Select


@dataclass
class NextPage:
    offset: int
    limit: int
    page: int
    next_page: int
    total: int
    total_pages: int


@dataclass
class Pagination:
    total: int
    limit: int = 100

    def calculate(self, page: int) -> NextPage:
        """ agnostic pagination, if it's the end it will returns
        -1 as the last page
        """
        total_pages = int(math.ceil(self.total/self.limit))
        _offset = self.limit * (page - 1)
        next_page = page + 1
        next_offset = self.limit * page
        if next_offset >= self.total:
            next_page = -1
        # print(
        #     f"limit: {self.limit}, offset: {_offset}, next_page: {next_page}")
        return NextPage(offset=_offset,
                        page=page,
                        next_page=next_page,
                        limit=self.limit,
                        total=self.total,
                        total_pages=total_pages)


async def get_total_async(session, Model):
    """Should be made in a context manager"""
    stmt = select(func.count(Model.id))
    _total = await session.execute(stmt)
    total = _total.scalar()

    return total


def get_total(session, Model):
    """Should be made in a context manager"""
    stmt = select(func.count(Model.id))
    _total = session.execute(stmt)
    total = _total.scalar()

    return total


def get_total_table(session, table: Table) -> int:
    stmt = select([func.count()]).select_from(table)
    res = session.execute(stmt).scalar()
    return res


async def get_total_table_async(session, table: Table) -> int:
    stmt = select([func.count()]).select_from(table)
    res = await session.execute(stmt)
    return res.scalar()
